'''
Routing for standard pipe connections

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from channels.routing import ProtocolTypeRouter, URLRouter
from channels.auth import AuthMiddlewareStack
from channels.http import AsgiHandler

from django.conf.urls import url
from django.urls import re_path

from .consumers import MessageConsumer, PokePipeConsumer
from .util import pipe_ws_endpoint_name, http_endpoint, http_poke_endpoint_enabled

# TODO document this and discuss embedding with other routes

http_routes = [
    ]

if http_poke_endpoint_enabled():
    http_routes.append(re_path(http_endpoint("poke"), PokePipeConsumer))

http_routes.append(re_path("^", AsgiHandler)) # AsgiHandler is 'the normal Django view handlers'

application = ProtocolTypeRouter({
    'websocket': AuthMiddlewareStack(URLRouter([re_path(pipe_ws_endpoint_name(), MessageConsumer),])),
    'http': AuthMiddlewareStack(URLRouter(http_routes)),
    })



'''
Utility functions

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''
import json
import uuid


from _plotly_utils.optional_imports import get_module


from django.conf import settings
from django.core.cache import cache
from django.utils.module_loading import import_string

from django_plotly_dash._patches import DjangoPlotlyJSONEncoder


def _get_settings():
    try:
        the_settings = settings.PLOTLY_DASH
    except AttributeError:
        the_settings = None

    return the_settings if the_settings else {}

def pipe_ws_endpoint_name():
    'Return the endpoint for pipe websocket connections'
    return _get_settings().get('ws_route', 'dpd/ws/channel')

def dpd_http_endpoint_root():
    'Return the root of the http endpoint for direct insertion of pipe messages'
    return _get_settings().get('http_route', 'dpd/views')

def http_endpoint(stem):
    'Form the http endpoint for a specific stem'
    return "^%s/%s/$" % (dpd_http_endpoint_root(), stem)

def insert_demo_migrations():
    'Check settings and report if objects for demo purposes should be inserted during migration'

    return _get_settings().get('insert_demo_migrations', False)

def http_poke_endpoint_enabled():
    'Return true if the http endpoint is enabled through the settings'
    return _get_settings().get('http_poke_enabled', True)

def cache_timeout_initial_arguments():
    'Return cache timeout, in seconds, for initial arguments'
    return _get_settings().get('cache_timeout_initial_arguments', 60)

def initial_argument_location():
    'Return True if cache to be used for setting and getting initial arguments, or False for a session'

    setget_location = _get_settings().get('cache_arguments', True)

    return setget_location

def store_initial_arguments(request, initial_arguments=None):
    'Store initial arguments, if any, and return a cache identifier'

    if initial_arguments is None:
        return None

    # convert to dict is json string
    if isinstance(initial_arguments, str):
        initial_arguments = json.loads(initial_arguments)

    # Generate a cache id
    cache_id = "dpd-initial-args-%s" % str(uuid.uuid4()).replace('-', '')

    # Store args in json form in cache
    if initial_argument_location():
        cache.set(cache_id, initial_arguments, cache_timeout_initial_arguments())
    else:
        request.session[cache_id] = initial_arguments

    return cache_id

def get_initial_arguments(request, cache_id=None):
    'Extract initial arguments for the dash app'

    if cache_id is None:
        return None

    if initial_argument_location():
        return cache.get(cache_id)

    return request.session[cache_id]

def static_asset_root():
    return _get_settings().get('static_asset_root','dpd/assets')

def full_asset_path(module_name, asset_path):
    path_contrib = "%s/%s/%s" %(static_asset_root(),
                                "/".join(module_name.split(".")),
                                asset_path)
    return path_contrib

def static_asset_path(module_name, asset_path):
    return static_path(full_asset_path(module_name, asset_path))

def serve_locally():
    return _get_settings().get('serve_locally', False)

def static_path(relative_path):
    try:
        static_url = settings.STATIC_URL
    except:
        static_url = '/static/'
    return "%s%s" %(static_url, relative_path)

def stateless_app_lookup_hook():
    'Return a function that performs lookup for aa stateless app, given its name, or returns None'

    func_name = _get_settings().get('stateless_loader', None)
    if func_name:
        func = import_string(func_name)
        return func

    # Default is no additional lookup
    return lambda _: None





'''
Django ORM models for dash applications

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import json
import logging

from django.db import models
from django.contrib import admin
from django.utils.text import slugify
from django.shortcuts import get_object_or_404

from .dash_wrapper import get_local_stateless_by_name, get_local_stateless_list, wid2str

logger = logging.getLogger(__name__)


def get_stateless_by_name(name):
    'Locate stateless app instance given its name'
    return get_local_stateless_by_name(name)


class StatelessApp(models.Model):
    '''
    A stateless Dash app. An instance of this model represents a dash app without any specific state
    '''
    app_name = models.CharField(max_length=100, blank=False, null=False, unique=True)
    slug = models.SlugField(max_length=110, unique=True, blank=True)

    def __str__(self):
        return self.app_name

    def save(self, *args, **kwargs): # pylint: disable=arguments-differ
        if not self.slug or len(self.slug) < 2:
            self.slug = slugify(self.app_name)
            exist_count = StatelessApp.objects.filter(slug__startswith=self.slug).count()
            if exist_count > 0:
                self.slug = self.slug + str(exist_count+1)
        return super().save(*args, **kwargs)

    def as_dash_app(self):
        '''
        Return a DjangoDash instance of the dash application
        '''
        dateless_dash_app = getattr(self, '_stateless_dash_app_instance', None)
        if not dateless_dash_app:
            dateless_dash_app = get_stateless_by_name(self.app_name)
            setattr(self, '_stateless_dash_app_instance', dateless_dash_app)
        return dateless_dash_app

def find_stateless_by_name(name):
    '''
    Find stateless app given its name

    First search the Django ORM, and if not found then look the app up in a local registry.
    If the app does not have an ORM entry then a StatelessApp model instance is created.
    '''
    try:
        dsa_app = StatelessApp.objects.get(app_name=name) # pylint: disable=no-member
        return dsa_app.as_dash_app()
    except: # pylint: disable=bare-except
        pass

    dash_app = get_stateless_by_name(name)
    dsa_app = StatelessApp(app_name=name)
    dsa_app.save()
    return dash_app


def check_stateless_loaded():
    for ua in get_local_stateless_list():
        try:
            find_stateless_by_name(ua)
        except:
            logger.warning("django-plotly-dash: Unable to create stateless instance: "+str(ua))


class StatelessAppAdmin(admin.ModelAdmin):
    'Admin for StatelessApp ORM model instances'
    list_display = ['app_name', 'slug',]
    list_filter = ['app_name', 'slug',]

    def check_registered(modeladmin, request, queryset):
        # Check all existing apps, keep if OK
        for sa in queryset.all():
            try:
                q = sa.as_dash_app()
            except:
                logger.warning("django-plotly-dash: Unable to load stateless app: "+str(sa))


    check_registered.short_description = "Check stateless apps"

    actions = [check_registered,]


class DashApp(models.Model):
    '''
    An instance of this model represents a dash application and its internal state
    '''
    stateless_app = models.ForeignKey(StatelessApp, on_delete=models.PROTECT,
                                      unique=False, null=False, blank=False)
    instance_name = models.CharField(max_length=100, unique=True,
                                     blank=True, null=False)
    slug = models.SlugField(max_length=110, unique=True, blank=True)

    # If mandating postgresql then this could be a JSONField
    base_state = models.TextField(null=False, default="{}")

    creation = models.DateTimeField(auto_now_add=True)
    update = models.DateTimeField(auto_now=True)
    save_on_change = models.BooleanField(null=False, default=False)

    def __str__(self):
        return self.instance_name

    def save(self, *args, **kwargs): # pylint: disable=arguments-differ
        if not self.instance_name:
            existing_count = DashApp.objects.all().count() # pylint: disable=no-member
            self.instance_name = "%s-%i" %(self.stateless_app.app_name, existing_count+1) # pylint: disable=no-member
        if not self.slug or len(self.slug) < 2:
            self.slug = slugify(self.instance_name)
        super().save(*args, **kwargs)

    def handle_current_state(self):
        '''
        Check to see if the current hydrated state and the saved state are different.

        If they are, then persist the current state in the database by saving the model instance.
        '''
        if getattr(self, '_current_state_hydrated_changed', False) and self.save_on_change:
            new_base_state = json.dumps(getattr(self, '_current_state_hydrated', {}))
            if new_base_state != self.base_state:
                self.base_state = new_base_state
                self.save()

    def have_current_state_entry(self, wid, key):
        'Return True if there is a cached current state for this app'
        cscoll = self.current_state()
        c_state = cscoll.get(wid2str(wid), {})
        return key in c_state

    def update_current_state(self, wid, key, value):
        '''
        Update current state with a (possibly new) value associated with key

        If the key does not represent an existing entry, then ignore it
        '''
        cscoll = self.current_state()
        c_state = cscoll.get(wid2str(wid), {})
        if key in c_state:
            current_value = c_state.get(key, None)
            if current_value != value:
                c_state[key] = value
                setattr(self, '_current_state_hydrated_changed', True)

    def current_state(self):
        '''
        Return the current internal state of the model instance.

        This is not necessarily the same as the persisted state
        stored in the self.base_state variable.
        '''
        c_state = getattr(self, '_current_state_hydrated', None)
        if not c_state:
            c_state = json.loads(self.base_state)
            setattr(self, '_current_state_hydrated', c_state)
            setattr(self, '_current_state_hydrated_changed', False)
        return c_state

    def as_dash_instance(self, cache_id=None):
        'Return a dash application instance for this model instance'
        dash_app = self.stateless_app.as_dash_app() # pylint: disable=no-member
        base = self.current_state()
        return dash_app.do_form_dash_instance(replacements=base,
                                              specific_identifier=self.slug,
                                              cache_id=cache_id)

    def _get_base_state(self):
        '''
        Get the base state of the object, as defined by the app.layout code, as a python dict
        '''
        base_app_inst = self.stateless_app.as_dash_app().as_dash_instance() # pylint: disable=no-member

        # Get base layout response, from a base object
        base_resp = base_app_inst.locate_endpoint_function('dash-layout')()

        base_obj = json.loads(base_resp.data.decode('utf-8'))

        # Walk the base layout and find all values; insert into base state map
        obj = {}
        base_app_inst.walk_tree_and_extract(base_obj, obj)
        return obj

    def populate_values(self):
        '''
        Add values from the underlying dash layout configuration
        '''
        obj = self._get_base_state()
        self.base_state = json.dumps(obj)

    @staticmethod
    def locate_item(ident, stateless=False, cache_id=None):
        '''Locate a dash application, given either the
        slug of an instance or the name for a stateless app'''
        if stateless:
            dash_app = find_stateless_by_name(ident)
        else:
            dash_app = get_object_or_404(DashApp, slug=ident)

        app = dash_app.as_dash_instance(cache_id=cache_id)
        return dash_app, app

class DashAppAdmin(admin.ModelAdmin):
    'Admin class for DashApp model'
    list_display = ['instance_name', 'stateless_app', 'slug',
                    'creation', 'update', 'save_on_change',]
    list_filter = ['creation', 'update', 'save_on_change', 'stateless_app',]

    def _populate_values(self, request, queryset): # pylint: disable=no-self-use, unused-argument
        for dash_app in queryset:
            dash_app.populate_values()
            dash_app.save()
    _populate_values.short_description = "Populate app instance"

    def _clone(self, request, queryset): # pylint: disable=no-self-use, unused-argument
        for dash_app in queryset:
            nda = DashApp(stateless_app=dash_app.stateless_app,
                          base_state=dash_app.base_state,
                          save_on_change=dash_app.save_on_change)
            nda.save()

    _clone.short_description = "Clone app instance"

    actions = ['_populate_values', '_clone',]



'''
Template tags for exposing dash applications in Django templates

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

# pylint: disable=too-many-arguments, unused-variable, unused-argument, possibly-unused-variable

from django import template

from django.contrib.sites.shortcuts import get_current_site

from django_plotly_dash.models import DashApp
from django_plotly_dash.util import pipe_ws_endpoint_name, store_initial_arguments


register = template.Library()


ws_default_url = "/%s" % pipe_ws_endpoint_name()


SANDBOX_SETTINGS = ["allow-downloads",
                    "allow-scripts",
                    "allow-same-origin",
                    "allow-modals",
                    "allow-popups",
                    "allow-popups-to-escape-sandbox",
                    ]


SANDBOX_STRING = " ".join(SANDBOX_SETTINGS)


def _locate_daapp(name, slug, da, cache_id=None):

    app = None

    if name is not None:
        da, app = DashApp.locate_item(name, stateless=True, cache_id=cache_id)

    if slug is not None:
        da, app = DashApp.locate_item(slug, stateless=False, cache_id=cache_id)

    if not app:
        app = da.as_dash_instance()

    return da, app

@register.inclusion_tag("django_plotly_dash/plotly_app.html", takes_context=True)
def plotly_app(context, name=None, slug=None, da=None, ratio=0.1, use_frameborder=False, initial_arguments=None):
    'Insert a dash application using a html iframe'

    fbs = '1' if use_frameborder else '0'

    dstyle = """
    position: relative;
    padding-bottom: %s%%;
    height: 0;
    overflow:hidden;
    """ % (ratio*100)

    istyle = """
    position: absolute;
    top: 0;
    left: 0;
    width: 100%;
    height: 100%;
    """

    cache_id = store_initial_arguments(context['request'], initial_arguments)

    da, app = _locate_daapp(name, slug, da, cache_id=cache_id)

    sandbox_settings = SANDBOX_STRING
    
    return locals()


@register.inclusion_tag("django_plotly_dash/plotly_app_bootstrap.html", takes_context=True)
def plotly_app_bootstrap(context, name=None, slug=None, da=None, aspect="4by3", initial_arguments=None):
    'Insert a dash application using a html iframe'

    valid_ratios = ['21by9',
                    '16by9',
                    '4by3',
                    '1by1',
                    ]

    if aspect not in valid_ratios:
        raise ValueError("plotly_app_bootstrap requires a valid aspect ratio from %s, but was supplied %s" % (str(valid_ratios),
                                                                                                              aspect))

    cache_id = store_initial_arguments(context['request'], initial_arguments)

    da, app = _locate_daapp(name, slug, da, cache_id=cache_id)

    sandbox_settings = SANDBOX_STRING
    
    return locals()


@register.simple_tag(takes_context=True)
def plotly_header(context):
    'Insert placeholder for django-plotly-dash header content'
    return context.request.dpd_content_handler.header_placeholder


@register.simple_tag(takes_context=True)
def plotly_footer(context):
    'Insert placeholder for django-plotly-dash footer content'
    return context.request.dpd_content_handler.footer_placeholder


@register.inclusion_tag("django_plotly_dash/plotly_direct.html", takes_context=True)
def plotly_direct(context, name=None, slug=None, da=None):
    'Direct insertion of a Dash app'

    da, app = _locate_daapp(name, slug, da)

    view_func = app.locate_endpoint_function()

    # Load embedded holder inserted by middleware
    eh = context.request.dpd_content_handler.embedded_holder

    # Need to add in renderer launcher
    renderer_launcher = '<script id="_dash-renderer" type="application/javascript">var renderer = new DashRenderer();</script>'

    app.set_embedded(eh)
    try:
        resp = view_func()
    finally:
        eh.add_scripts(renderer_launcher)
        app.exit_embedded()

    return locals()


@register.inclusion_tag("django_plotly_dash/plotly_messaging.html", takes_context=True)
def plotly_message_pipe(context, url=None):
    'Insert script for providing background websocket connection'
    url = url if url else ws_default_url
    return locals()


@register.simple_tag()
def plotly_app_identifier(name=None, slug=None, da=None, postfix=None):
    'Return a slug-friendly identifier'

    da, app = _locate_daapp(name, slug, da)

    slugified_id = app.slugified_id()

    if postfix:
        return "%s-%s" %(slugified_id, postfix)
    return slugified_id


@register.simple_tag()
def plotly_class(name=None, slug=None, da=None, prefix=None, postfix=None, template_type=None):
    'Return a string of space-separated class names'

    da, app = _locate_daapp(name, slug, da)

    return app.extra_html_properties(prefix=prefix,
                                     postfix=postfix,
                                     template_type=template_type)


@register.simple_tag(takes_context=True)
def site_root_url(context):
    'Provide the root url of the demo site'
    current_site_url = get_current_site(context.request)
    return current_site_url.domain



'''
dash_wrapper

This module provides a DjangoDash class that can be used to
expose a Plotly Dasb application through a Django server

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''
import inspect
import itertools
import json
import warnings
from typing import Dict, List, Callable

import dash
from dash import Dash, dependencies
from dash._utils import split_callback_id, inputs_to_dict
from django.urls import reverse
from django.utils.text import slugify
from flask import Flask

from .app_name import app_name, main_view_label
from .middleware import EmbeddedHolder
from .util import serve_locally as serve_locally_setting
from .util import stateless_app_lookup_hook
from .util import static_asset_path, DjangoPlotlyJSONEncoder

try:
    from dataclasses import dataclass

    @dataclass(frozen=True)
    class CallbackContext:
        inputs_list : List
        inputs: Dict
        states_list: List
        states: Dict
        outputs_list: List
        outputs: Dict
        triggered: List

except:
    # Not got python 3.7 or dataclasses yet
    class CallbackContext:
        def __init__(self, **kwargs):
            self._args = kwargs

        @property
        def inputs_list(self):
            return self._args['inputs_list']

        @property
        def inputs(self):
            return self._args['inputs']

        @property
        def states_list(self):
            return self._args['states_list']

        @property
        def states(self):
            return self._args['states']

        @property
        def outputs(self):
            return self._args['outputs']

        @property
        def outputs_list(self):
            return self._args['outputs_list']
        @property
        def triggered(self):
            return self._args['triggered']


uid_counter = 0

usable_apps = {}

_stateless_app_lookup_func = None

def add_usable_app(name, app):
    'Add app to local registry by name'
    global usable_apps # pylint: disable=global-statement
    usable_apps[name] = app
    return name

def all_apps():
    'Return a dictionary of all locally registered apps with the slug name as key'
    return usable_apps

def get_local_stateless_list():
    """Return a list of all locally registered stateless apps
    """
    return list(usable_apps)

def get_local_stateless_by_name(name):
    '''
    Locate a registered dash app by name, and return a DjangoDash instance encapsulating the app.
    '''
    sa = usable_apps.get(name, None)

    if not sa:

        global _stateless_app_lookup_func # pylint: disable=global-statement

        if _stateless_app_lookup_func is None:
            _stateless_app_lookup_func = stateless_app_lookup_hook()

        sa = _stateless_app_lookup_func(name)

    if not sa:
        # TODO wrap this in raising a 404 if not found
        raise KeyError("Unable to find stateless DjangoApp called %s"%name)

    return sa


class Holder:
    'Helper class for holding configuration options'
    def __init__(self):
        self.items = []
    def append_css(self, stylesheet):
        'Add extra css file name to component package'
        self.items.append(stylesheet)
    def append_script(self, script):
        'Add extra script file name to component package'
        self.items.append(script)


class DjangoDash:
    '''
    Wrapper class that provides Dash functionality in a form that can be served by Django

    To use, construct an instance of DjangoDash() in place of a Dash() one.
    '''
    #pylint: disable=too-many-instance-attributes
    def __init__(self, name=None, serve_locally=None,
                 add_bootstrap_links=False,
                 suppress_callback_exceptions=False,
                 external_stylesheets=None,
                 external_scripts=None,
                 **kwargs):  # pylint: disable=unused-argument, too-many-arguments

        # store arguments to pass them later to the WrappedDash instance
        self.external_stylesheets = external_stylesheets or []
        self.external_scripts = external_scripts or []
        self._kwargs = kwargs
        if kwargs:
            warnings.warn("You are passing extra arguments {kwargs} that will be passed to Dash(...) "
                          "but may not be properly handled by django-plotly-dash.".format(kwargs=kwargs))

        if name is None:
            global uid_counter # pylint: disable=global-statement
            uid_counter += 1
            self._uid = "djdash_%i" % uid_counter
        else:
            self._uid = name
        self.layout = None
        self._callback_sets = []
        self._clientside_callback_sets = []

        self.css = Holder()
        self.scripts = Holder()

        add_usable_app(self._uid,
                       self)

        if serve_locally is None:
            self._serve_locally = serve_locally_setting()
        else:
            self._serve_locally = serve_locally

        self._suppress_callback_exceptions = suppress_callback_exceptions

        if add_bootstrap_links:
            from bootstrap4.bootstrap import css_url
            bootstrap_source = css_url()['href']

            if self._serve_locally:
                # Ensure package is loaded; if not present then pip install dpd-static-support
                hard_coded_package_name = "dpd_static_support"
                base_file_name = bootstrap_source.split('/')[-1]

                self.css.append_script({'external_url':        [bootstrap_source,],
                                        'relative_package_path' : base_file_name,
                                        'namespace':              hard_coded_package_name,
                                        })
            else:
                self.css.append_script({'external_url':[bootstrap_source,],})

        # Remember some caller info for static files
        caller_frame = inspect.stack()[1]
        self.caller_module = inspect.getmodule(caller_frame[0])
        try:
            self.caller_module_location = inspect.getfile(self.caller_module)
        except:
            self.caller_module_location = None
        self.assets_folder = "assets"

    def get_asset_static_url(self, asset_path):
        module_name = self.caller_module.__name__
        return static_asset_path(module_name, asset_path)

    def as_dash_instance(self, cache_id=None):
        '''
        Form a dash instance, for stateless use of this app
        '''
        return self.do_form_dash_instance(cache_id=cache_id)

    def handle_current_state(self):
        'Do nothing impl - only matters if state present'
        pass
    def update_current_state(self, wid, key, value):
        'Do nothing impl - only matters if state present'
        pass
    def have_current_state_entry(self, wid, key):
        'Do nothing impl - only matters if state present'
        pass

    def get_base_pathname(self, specific_identifier, cache_id):
        'Base path name of this instance, taking into account any state or statelessness'
        if not specific_identifier:
            app_pathname = "%s:app-%s"% (app_name, main_view_label)
            ndid = self._uid
        else:
            app_pathname = "%s:%s" % (app_name, main_view_label)
            ndid = specific_identifier

        kwargs = {'ident': ndid}

        if cache_id:
            kwargs['cache_id'] = cache_id
            app_pathname = app_pathname + "--args"

        full_url = reverse(app_pathname, kwargs=kwargs)
        if full_url[-1] != '/':
            full_url = full_url + '/'
        return ndid, full_url

    def do_form_dash_instance(self, replacements=None, specific_identifier=None, cache_id=None):
        'Perform the act of constructing a Dash instance taking into account state'

        ndid, base_pathname = self.get_base_pathname(specific_identifier, cache_id)
        return self.form_dash_instance(replacements, ndid, base_pathname)

    def form_dash_instance(self, replacements=None, ndid=None, base_pathname=None):
        'Construct a Dash instance taking into account state'

        if ndid is None:
            ndid = self._uid

        rd = WrappedDash(base_pathname=base_pathname,
                         replacements=replacements,
                         ndid=ndid,
                         serve_locally=self._serve_locally,
                         external_stylesheets=self.external_stylesheets,
                         external_scripts=self.external_scripts,
                         **self._kwargs)

        rd.layout = self.layout
        rd.config['suppress_callback_exceptions'] = self._suppress_callback_exceptions

        for cb, func in self._callback_sets:
            rd.callback(**cb)(func)
        for cb in self._clientside_callback_sets:
            rd.clientside_callback(**cb)
        for s in self.css.items:
            rd.css.append_css(s)
        for s in self.scripts.items:
            rd.scripts.append_script(s)

        return rd

    @staticmethod
    def get_expanded_arguments(func, inputs, state):
        """Analyse a callback function signature to detect the expanded arguments to add when called.
        It uses the inputs and the state information to identify what arguments are already coming from Dash.

        It returns a list of the expanded parameters to inject (can be [] if nothing should be injected)
         or None if all parameters should be injected."""
        n_dash_parameters = len(inputs or []) + len(state or [])

        parameter_types = {kind: [p.name for p in parameters] for kind, parameters in
                           itertools.groupby(inspect.signature(func).parameters.values(), lambda p: p.kind)}
        if inspect.Parameter.VAR_KEYWORD in parameter_types:
            # there is some **kwargs, inject all parameters
            expanded = None
        elif inspect.Parameter.VAR_POSITIONAL in parameter_types:
            # there is a *args, assume all parameters afterwards (KEYWORD_ONLY) are to be injected
            # some of these parameters may not be expanded arguments but that is ok
            expanded = parameter_types.get(inspect.Parameter.KEYWORD_ONLY, [])
        else:
            # there is no **kwargs, filter argMap to take only the keyword arguments
            expanded = parameter_types.get(inspect.Parameter.POSITIONAL_OR_KEYWORD, [])[
                       n_dash_parameters:] + parameter_types.get(inspect.Parameter.KEYWORD_ONLY, [])

        return expanded

    def callback(self, *_args, **_kwargs):
        '''Form a callback function by wrapping, in the same way as the underlying Dash application would
        but handling extra arguments provided by dpd.

        It will inspect the signature of the function to ensure only relevant expanded arguments are passed to the callback.

        If the function accepts a **kwargs => all expanded arguments are sent to the function in the kwargs.
        If the function has a *args => expanded arguments matching parameters after the *args are injected.
        Otherwise, take all arguments beyond the one provided by Dash (based on the Inputs/States provided).

        '''

        output, inputs, state, prevent_initial_call = dependencies.handle_callback_args(
            _args, _kwargs
        )

        callback_set = {'output': output,
                        'inputs': inputs,
                        'state': state,
                        'prevent_initial_call': prevent_initial_call}

        def wrap_func(func):
            self._callback_sets.append((callback_set, func))
            # add an expanded attribute to the function with the information to use in dispatch_with_args
            # to inject properly only the expanded arguments the function can accept
            # if .expanded is None => inject all
            # if .expanded is a list => inject only
            func.expanded = DjangoDash.get_expanded_arguments(func, inputs, state)
            return func
        return wrap_func

    expanded_callback = callback

    def clientside_callback(self, clientside_function,  *_args, **_kwargs):
        'Form a callback function by wrapping, in the same way as the underlying Dash application would'
        output, inputs, state, prevent_initial_call = dependencies.handle_callback_args(
            _args, _kwargs
        )

        callback_set = { 'clientside_function': clientside_function,
                         'output': output,
                         'inputs': inputs,
                         'state': state,
                         'prevent_initial_call': prevent_initial_call}

        self._clientside_callback_sets.append(callback_set)


    def get_asset_url(self, asset_name):
        '''URL of an asset associated with this component

        Use a placeholder and insert later
        '''

        return "assets/" + str(asset_name)

        #return self.as_dash_instance().get_asset_url(asset_name)

class PseudoFlask(Flask):
    'Dummy implementation of a Flask instance, providing stub functionality'
    def __init__(self):
        self.config = {'DEBUG': False}
        self.endpoints = {}
        self.name = "PseudoFlaskDummyName"
        self.blueprints = {}
        self._got_first_request = False
        self.before_request_funcs = {}

    # pylint: disable=unused-argument, missing-docstring

    def after_request(self, *args, **kwargs):
        pass
    def errorhandler(self, *args, **kwargs): # pylint: disable=no-self-use
        def eh_func(f):
            return args[0]
        return eh_func
    def add_url_rule(self, *args, **kwargs):
        route = kwargs['endpoint']
        self.endpoints[route] = kwargs
    def before_first_request(self, *args, **kwargs):
        pass
    def run(self, *args, **kwargs):
        pass
    def register_blueprint(self, *args, **kwargs):
        pass


def wid2str(wid):
    """Convert an python id (str or dict) into its Dash representation.

    see https://github.com/plotly/dash/blob/c5ba38f0ae7b7f8c173bda10b4a8ddd035f1d867/dash-renderer/src/actions/dependencies.js#L114"""
    if isinstance(wid, str):
        return wid
    data = ",".join(f"{json.dumps(k)}:{json.dumps(v)}" for k, v in sorted(wid.items()))
    return f"{{{data}}}"


class WrappedDash(Dash):
    'Wrapper around the Plotly Dash application instance'
    # pylint: disable=too-many-arguments, too-many-instance-attributes
    def __init__(self,
                 base_pathname=None, replacements=None, ndid=None, serve_locally=False,
                 **kwargs):

        self._uid = ndid

        self._flask_app = Flask(self._uid)
        self._notflask = PseudoFlask()
        self._base_pathname = base_pathname

        kwargs['url_base_pathname'] = self._base_pathname
        kwargs['server'] = self._notflask

        #xkwargs['DEBUG'] = kwargs.get('DEBUG', False)

        super().__init__(__name__, **kwargs)

        self.css.config.serve_locally = serve_locally
        self.scripts.config.serve_locally = serve_locally

        self._adjust_id = False
        if replacements:
            self._replacements = replacements
        else:
            self._replacements = dict()
        self._use_dash_layout = len(self._replacements) < 1

        self._return_embedded = False

    def use_dash_dispatch(self):
        """Return True if underlying dash dispatching should be used.

        This stub is present to allow older code to work. Following PR #304
        (see https://github.com/GibbsConsulting/django-plotly-dash/pull/304/files for
        details) this function is no longer needed and therefore should always
        return False"""
        return False

    def use_dash_layout(self):
        '''
        Indicate if the underlying dash layout can be used.

        If application state is in use, then the underlying dash layout functionality has to be
        augmented with the state information and this function returns False
        '''
        return self._use_dash_layout

    def augment_initial_layout(self, base_response, initial_arguments=None):
        'Add application state to initial values'
        if self.use_dash_layout() and not initial_arguments and False:
            return base_response.data, base_response.mimetype

        # Adjust the base layout response
        baseDataInBytes = base_response.data
        baseData = json.loads(baseDataInBytes.decode('utf-8'))

        # Also add in any initial arguments
        if not initial_arguments:
            initial_arguments = {}

        # Define overrides as self._replacements updated with initial_arguments
        overrides = dict(self._replacements)
        overrides.update(initial_arguments)

        # Walk tree. If at any point we have an element whose id
        # matches, then replace any named values at this level
        reworked_data = self.walk_tree_and_replace(baseData, overrides)

        response_data = json.dumps(reworked_data,
                                   cls=DjangoPlotlyJSONEncoder)

        return response_data, base_response.mimetype

    def walk_tree_and_extract(self, data, target):
        'Walk tree of properties and extract identifiers and associated values'
        if isinstance(data, dict):
            for key in ['children', 'props']:
                self.walk_tree_and_extract(data.get(key, None), target)
            ident = data.get('id', None)
            if ident is not None:
                ident = wid2str(ident)
                idVals = target.get(ident, {})
                for key, value in data.items():
                    if key not in ['props', 'options', 'children', 'id']:
                        idVals[key] = value
                if idVals:
                    target[ident] = idVals
        if isinstance(data, list):
            for element in data:
                self.walk_tree_and_extract(element, target)

    def walk_tree_and_replace(self, data, overrides):
        '''
        Walk the tree. Rely on json decoding to insert instances of dict and list
        ie we use a dna test for anatine, rather than our eyes and ears...
        '''
        if isinstance(data, dict):
            response = {}
            replacements = {}
            # look for id entry
            thisID = data.get('id', None)
            if isinstance(thisID, dict):
                # handle case of thisID being a dict (pattern) => linear search in overrides dict
                thisID = wid2str(thisID)
                for k, v in overrides.items():
                    if thisID == k:
                        replacements = v
                        break
            elif thisID is not None:
                # handle standard case of string thisID => key lookup
                replacements = overrides.get(thisID, {})
            # walk all keys and replace if needed
            for k, v in data.items():
                r = replacements.get(k, None)
                if r is None:
                    r = self.walk_tree_and_replace(v, overrides)
                response[k] = r
            return response
        if isinstance(data, list):
            # process each entry in turn and return
            return [self.walk_tree_and_replace(x, overrides) for x in data]
        return data

    def flask_app(self):
        'Underlying flask application for stub implementation'
        return self._flask_app

    def base_url(self):
        'Base url of this component'
        return self._base_pathname

    def app_context(self, *args, **kwargs):
        'Extract application context from underlying flask application'
        return self._flask_app.app_context(*args,
                                           **kwargs)

    def test_request_context(self, *args, **kwargs):
        'Request context for testing from underluying flask application'
        return self._flask_app.test_request_context(*args,
                                                    **kwargs)

    def locate_endpoint_function(self, name=None):
        'Locate endpoint function given name of view'
        if name is not None:
            ep = "%s_%s" %(self._base_pathname,
                           name)
        else:
            ep = self._base_pathname
        return self._notflask.endpoints[ep]['view_func']

    # pylint: disable=no-member
    @Dash.layout.setter
    def layout(self, value):
        'Overloaded layout function to fix component names as needed'

        if self._adjust_id:
            self._fix_component_id(value)
        return Dash.layout.fset(self, value)

    def _fix_component_id(self, component):
        'Fix name of component ad all of its children'

        theID = getattr(component, "id", None)
        if theID is not None:
            setattr(component, "id", self._fix_id(theID))
        try:
            for c in component.children:
                self._fix_component_id(c)
        except: #pylint: disable=bare-except
            pass

    def _fix_id(self, name):
        'Adjust identifier to include component name'
        if not self._adjust_id:
            return name
        return "%s_-_%s" %(self._uid,
                           name)

    def _fix_callback_item(self, item):
        'Update component identifier'
        item.component_id = self._fix_id(item.component_id)
        return item

    def callback(self, output, inputs, state, prevent_initial_call):
        'Invoke callback, adjusting variable names as needed'

        if isinstance(output, (list, tuple)):
            fixed_outputs = [self._fix_callback_item(x) for x in output]
        else:
            fixed_outputs = self._fix_callback_item(output)

        return super().callback(fixed_outputs,
                                [self._fix_callback_item(x) for x in inputs],
                                [self._fix_callback_item(x) for x in state],
                                prevent_initial_call=prevent_initial_call)

    def clientside_callback(self, clientside_function, output, inputs, state, prevent_initial_call): # pylint: disable=dangerous-default-value
        'Invoke callback, adjusting variable names as needed'

        if isinstance(output, (list, tuple)):
            fixed_outputs = [self._fix_callback_item(x) for x in output]
        else:
            fixed_outputs = self._fix_callback_item(output)

        return super().clientside_callback(clientside_function,
                                           fixed_outputs,
                                           [self._fix_callback_item(x) for x in inputs],
                                           [self._fix_callback_item(x) for x in state],
                                           prevent_initial_call=prevent_initial_call)

    #pylint: disable=too-many-locals
    def dispatch_with_args(self, body, argMap):
        'Perform callback dispatching, with enhanced arguments and recording of response'
        inputs = body.get('inputs', [])
        input_values = inputs_to_dict(inputs)
        states = body.get('state', [])
        output = body['output']
        outputs_list = body.get('outputs') or split_callback_id(output)
        changed_props = body.get('changedPropIds', [])
        triggered_inputs = [{"prop_id": x, "value": input_values.get(x)} for x in changed_props]

        callback_context_info = {
            'inputs_list': inputs,
            'inputs': input_values,
            'states_list': states,
            'states': inputs_to_dict(states),
            'outputs_list': outputs_list,
            'outputs': outputs_list,
            'triggered': triggered_inputs,
            }

        callback_context = CallbackContext(**callback_context_info)

        # Overload dash global variable
        dash.callback_context = callback_context

        # Add context to arg map, if extended callbacks in use
        if len(argMap) > 0:
            argMap['callback_context'] = callback_context

        single_case = not(output.startswith('..') and output.endswith('..'))
        if single_case:
            # single Output (not in a list)
            outputs = [output]
        else:
            # multiple outputs in a list (the list could contain a single item)
            outputs = output[2:-2].split('...')


        da = argMap.get('dash_app', None)

        callback_info = self.callback_map[output]

        args = []

        for c in inputs + states:
            if isinstance(c, list):  # ALL, ALLSMALLER
                v = [ci.get("value") for ci in c]
                if da:
                    for ci, vi in zip(c, v):
                        da.update_current_state(ci['id'], ci['property'], vi)
            else:
                v = c.get("value")
                if da:
                    da.update_current_state(c['id'], c['property'], v)

            args.append(v)


        # Dash 1.11 introduces a set of outputs
        outputs_list = body.get('outputs') or split_callback_id(output)
        argMap['outputs_list'] = outputs_list

        # Special: intercept case of insufficient arguments
        # This happens when a property has been updated with a pipe component
        # TODO see if this can be attacked from the client end

        if len(args) < len(callback_info['inputs']):
            return 'EDGECASEEXIT'

        callback = callback_info["callback"]
        # smart injection of parameters if .expanded is defined
        if callback.expanded is not None:
            parameters_to_inject = {*callback.expanded, 'outputs_list'}
            res = callback(*args, **{k: v for k, v in argMap.items() if k in parameters_to_inject})
        else:
            res = callback(*args, **argMap)

        if da:
            class LazyJson:
                """A class to allow delayed the evaluation of a dict (returned by `func`)
                 till the first get(...) is called on the dict."""

                def __init__(self, func):
                    self._root_value = func

                def get(self, item, default):
                    if isinstance(self._root_value, Callable):
                        self._root_value = self._root_value()
                    return self._root_value.get(item, default)

            # wraps the json parsing of the response into LazyJson to avoid unnecessary parsing
            root_value = LazyJson(lambda: json.loads(res).get('response', {}))

            for output_item in outputs:
                if isinstance(output_item, str):
                    output_id, output_property = output_item.split('.')
                    if da.have_current_state_entry(output_id, output_property):
                        value = root_value.get(output_id,{}).get(output_property, None)
                        da.update_current_state(output_id, output_property, value)
                else:
                    # todo: implement saving of state for pattern matching ouputs
                    raise NotImplementedError("Updating state for dict keys (pattern matching) is not yet implemented")

        return res

    def slugified_id(self):
        'Return the app id in a slug-friendly form'
        pre_slugified_id = self._uid
        return slugify(pre_slugified_id)

    def extra_html_properties(self, prefix=None, postfix=None, template_type=None):
        '''
        Return extra html properties to allow individual apps to be styled separately.

        The content returned from this function is injected unescaped into templates.
        '''

        prefix = prefix if prefix else "django-plotly-dash"

        post_part = "-%s" % postfix if postfix else ""
        template_type = template_type if template_type else "iframe"

        slugified_id = self.slugified_id()

        return "%(prefix)s %(prefix)s-%(template_type)s %(prefix)s-app-%(slugified_id)s%(post_part)s" % {'slugified_id':slugified_id,
                                                                                                         'post_part':post_part,
                                                                                                         'template_type':template_type,
                                                                                                         'prefix':prefix,
                                                                                                        }

    def index(self, *args, **kwargs):  # pylint: disable=unused-argument
        scripts = self._generate_scripts_html()
        css = self._generate_css_dist_html()
        config = self._generate_config_html()
        metas = self._generate_meta_html()
        renderer = self._generate_renderer()
        title = getattr(self, 'title', 'Dash')
        if self._favicon:
            import flask
            favicon = '<link rel="icon" type="image/x-icon" href="{}">'.format(
                flask.url_for('assets.static', filename=self._favicon))
        else:
            favicon = ''

            _app_entry = '''
<div id="react-entry-point">
  <div class="_dash-loading">
    Loading...
  </div>
</div>
'''
        index = self.interpolate_index(
            metas=metas, title=title, css=css, config=config,
            scripts=scripts, app_entry=_app_entry, favicon=favicon,
            renderer=renderer)

        return index

    def interpolate_index(self, **kwargs): #pylint: disable=arguments-differ

        if not self._return_embedded:
            resp = super().interpolate_index(**kwargs)
            return resp

        self._return_embedded.add_css(kwargs['css'])
        self._return_embedded.add_config(kwargs['config'])
        self._return_embedded.add_scripts(kwargs['scripts'])

        return kwargs['app_entry']

    def set_embedded(self, embedded_holder=None):
        'Set a handler for embedded references prior to evaluating a view function'
        self._return_embedded = embedded_holder if embedded_holder else EmbeddedHolder()

    def exit_embedded(self):
        'Exit the embedded section after processing a view'
        self._return_embedded = False



# Generated by Django 2.0.5 on 2018-06-08 20:03

# pylint: skip-file

from django.db import migrations
from ..util import insert_demo_migrations

def addExamples(apps, schema_editor):

    if not insert_demo_migrations():
        return

    DashApp = apps.get_model("django_plotly_dash","DashApp")
    StatelessApp = apps.get_model("django_plotly_dash","StatelessApp")

    sa1 = StatelessApp(app_name="SimpleExample",
                       slug="simple-example")

    sa1.save()

    da1 = DashApp(stateless_app=sa1,
                  instance_name="SimpleExample-1",
                  slug="simpleexample-1",
                  base_state='{"dropdown-color":{"value":"blue"},"dropdown-size":{"value":"small"}}')

    da1.save()

    sa2 = StatelessApp(app_name="LiveOutput",
                       slug="liveoutput")

    sa2.save()

    sa21 = StatelessApp(app_name="LiveInput",
                       slug="liveinput")

    sa21.save()

    da2 = DashApp(stateless_app=sa2,
                  instance_name="liveoutput-2",
                  slug="liveoutput-2",
                  base_state='''{"named_count_pipe": {
                      "label": "named_counts",
                      "channel_name": "live_button_counter",
                      "value": null},
 "state_uid": {"value": "635c89ae-c996-4488-84df-041c1153fdef"}}
''',
                  save_on_change=True)

    da2.save()

    sa3 = StatelessApp(app_name="Ex2",
                       slug="ex2")

    sa3.save()

    da3 = DashApp(stateless_app=sa3,
                  instance_name="Ex2-1",
                  slug="ex2-3",
                  base_state='{"dropdown-one":{"value":"Nitrogen"}}',
                  save_on_change=True)

    da3.save()

    sa4 = StatelessApp(app_name="MultipleCallbackValues",
                       slug="multiple-callback-values")

    sa4.save()

    da4 = DashApp(stateless_app=sa4,
                  instance_name="Multiple Callback Values Example 1",
                  slug="multiple-callback-values-1")

    da4.save()

    sa5 = StatelessApp(app_name="MultipleCallbackValuesExpanded",
                       slug="multiple-callback-values-exapnded")

    sa5.save()

    da5 = DashApp(stateless_app=sa5,
                  instance_name="Multiple Callback Values Example 2",
                  slug="multiple-callback-values-expanded")

    da5.save()

    sa6 = StatelessApp(app_name="PatternStateCallbacks",
                       slug="pattern-state-callback")

    sa6.save()

    da6 = DashApp(stateless_app=sa6,
                  instance_name="Pattern and State saving Example",
                  slug="pattern-state-callback")

    da6.save()


def remExamples(apps, schema_editor):

    DashApp = apps.get_model("django_plotly_dash","DashApp")
    StatelessApp = apps.get_model("django_plotly_dash","StatelessApp")

    DashApp.objects.all().delete()
    StatelessApp.objects.all().delete()

class Migration(migrations.Migration):

    dependencies = [
        ('django_plotly_dash', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(addExamples, remExamples),
    ]



'''
Collation of patches made to other python libraries, such as Dash itself.

If/when the patches are not needed they will be removed from this file.


Copyright (c) 2022 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

import json


from plotly.io._json import config, clean_to_json_compatible
from plotly.utils import PlotlyJSONEncoder

from _plotly_utils.optional_imports import get_module
from django.utils.encoding import force_str
from django.utils.functional import Promise


class DjangoPlotlyJSONEncoder(PlotlyJSONEncoder):
    """Augment the PlotlyJSONEncoder class with Django delayed processing"""
    def default(self, obj):
        if isinstance(obj, Promise):
            return force_str(obj)
        return super().default(obj)


def promise_clean_to_json_compatible(obj):

    if isinstance(obj, dict):
        return {promise_clean_to_json_compatible(k): promise_clean_to_json_compatible(v) for k, v in obj.items()}

    if isinstance(obj, (list, tuple)):
        if obj:
            return [promise_clean_to_json_compatible(v) for v in obj]

    if isinstance(obj, Promise):
        return force_str(obj)

    #try:
    #    return obj.to_plotly_json()
    #except AttributeError:
    #    pass

    return obj


def to_json_django_plotly(plotly_object, pretty=False, engine=None):
    """
    Convert a plotly/Dash object to a JSON string representation

    Parameters
    ----------
    plotly_object:
        A plotly/Dash object represented as a dict, graph_object, or Dash component

    pretty: bool (default False)
        True if JSON representation should be pretty-printed, False if
        representation should be as compact as possible.

    engine: str (default None)
        The JSON encoding engine to use. One of:
          - "json" for an engine based on the built-in Python json module
          - "orjson" for a faster engine that requires the orjson package
          - "auto" for the "orjson" engine if available, otherwise "json"
        If not specified, the default engine is set to the current value of
        plotly.io.json.config.default_engine.

    Returns
    -------
    str
        Representation of input object as a JSON string

    See Also
    --------
    to_json : Convert a plotly Figure to JSON with validation
    """
    orjson = get_module("orjson", should_load=True)

    # Determine json engine
    if engine is None:
        engine = config.default_engine

    if engine == "auto":
        if orjson is not None:
            engine = "orjson"
        else:
            engine = "json"
    elif engine not in ["orjson", "json"]:
        raise ValueError("Invalid json engine: %s" % engine)

    modules = {
        "sage_all": get_module("sage.all", should_load=False),
        "np": get_module("numpy", should_load=False),
        "pd": get_module("pandas", should_load=False),
        "image": get_module("PIL.Image", should_load=False),
    }

    # Dump to a JSON string and return
    # --------------------------------
    if engine == "json":
        opts = {}
        if pretty:
            opts["indent"] = 2
        else:
            # Remove all whitespace
            opts["separators"] = (",", ":")

        return json.dumps(plotly_object, cls=DjangoPlotlyJSONEncoder, **opts)
    elif engine == "orjson":
        config.validate_orjson()
        opts = orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY

        if pretty:
            opts |= orjson.OPT_INDENT_2

        # Plotly

        try:
            plotly_object = plotly_object.to_plotly_json()
        except AttributeError:
            pass

        # Try without cleaning
        try:
            return orjson.dumps(plotly_object, option=opts).decode("utf8")
        except TypeError:
            pass

        cleaned = clean_to_json_compatible(
            plotly_object,
            numpy_allowed=True,
            datetime_allowed=True,
            modules=modules,
        )

        cleaned = promise_clean_to_json_compatible(cleaned)

        return orjson.dumps(cleaned, option=opts).decode("utf8")

import plotly.io.json
plotly.io.json.to_json_plotly = to_json_django_plotly



'''
Control url access based on user and other state

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import importlib
from django.conf import settings

from django.contrib.auth.decorators import login_required as login_required_decorator

#pylint: disable=bare-except, unused-argument

def login_required(view_function, **kwargs):
    'Wrap all DPD calls with login_required'
    return login_required_decorator(view_function)

try:
    dash_view_decorator_name = settings.PLOTLY_DASH['view_decorator']
    try:
        dash_view_decorator = locals()[dash_view_decorator_name]
    except:
        mod_name, func_name = dash_view_decorator_name.rsplit('.', 1)
        if mod_name:
            mod = importlib.import_module(mod_name)
            dash_view_decorator = getattr(mod, func_name)
        else:
            dash_view_decorator = locals()[func_name]
except:
    dash_view_decorator = None

def process_view_function(view_function, **kwargs):
    'Process view function and wrap according to settings'

    if dash_view_decorator:
        return dash_view_decorator(view_function, **kwargs)

    return view_function



'''
Staticfiles finders for Dash assets

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
import importlib

from collections import OrderedDict

from django.contrib.staticfiles.finders import BaseFinder
from django.contrib.staticfiles.utils import get_files

from django.core.files.storage import FileSystemStorage

from django.conf import settings
from django.apps import apps

from django_plotly_dash.dash_wrapper import all_apps
from django_plotly_dash.util import full_asset_path

class DashComponentFinder(BaseFinder):
    'Find static files in components'

    #pylint: disable=abstract-method, redefined-builtin

    def __init__(self):

        self.locations = []
        self.storages = OrderedDict()
        self.components = {}

        self.ignore_patterns = ["*.py", "*.pyc",]

        try:
            components = settings.PLOTLY_COMPONENTS
        except:
            components = []

        built_ins = [('dash', ['dcc', 'html', 'dash_table', 'deps', 'dash-renderer', 'dash-renderer/build']), ]

        for component_name in components:

            split_name = component_name.split('/')
            try:
                module_name = ".".join(split_name)
                module = importlib.import_module(module_name)
                path_directory = os.path.dirname(module.__file__)
            except:
                module_name = ".".join(split_name[:-1])
                module = importlib.import_module(module_name)
                path_directory = os.path.join(os.path.dirname(module.__file__),
                                              split_name[-1])

            root = path_directory
            storage = FileSystemStorage(location=root)
            path = "dash/component/%s" % component_name

            # Path_directory is where from
            # path is desired url mount point of the content of path_directory
            # component_name is the name of the component

            storage.prefix = path

            self.locations.append(component_name)

            self.storages[component_name] = storage
            self.components[path] = component_name

        for module_name, component_list in built_ins:

            module = importlib.import_module(module_name)

            for specific_component in component_list:

                path_directory = os.path.join(os.path.dirname(module.__file__),
                                              specific_component)

                root = path_directory
                component_name = f"{module_name}/{specific_component}"
                path = f"dash/component/{component_name}"

                if path not in self.components:

                    storage = FileSystemStorage(location=root)
                    storage.prefix = path

                    self.locations.append(component_name)

                    self.storages[component_name] = storage
                    self.components[path] = component_name

        super().__init__()

    def find(self, path, all=False):
        matches = []
        for component_name in self.locations:
            storage = self.storages[component_name]
            location = storage.location # dir on disc

            component_path = "dash/component/%s" % component_name
            if len(path) > len(component_path) and path[:len(component_path)] == component_path:

                matched_path = os.path.join(location, path[len(component_path)+1:])
                if os.path.exists(matched_path):
                    if not all:
                        return matched_path
                    matches.append(matched_path)

        return matches

    # pylint: disable=inconsistent-return-statements, no-self-use
    def find_location(self, path):
        'Return location, if it exists'
        if os.path.exists(path):
            return path

    def list(self, ignore_patterns):
        for component_name in self.locations:
            storage = self.storages[component_name]
            for path in get_files(storage, ignore_patterns + self.ignore_patterns):
                yield path, storage

class DashAppDirectoryFinder(BaseFinder):
    'Find static fies in application subdirectories'

    def __init__(self):
        # get all registered apps

        self.locations = []
        self.storages = OrderedDict()

        self.ignore_patterns = ["*.py", "*.pyc",]

        for app_config in apps.get_app_configs():

            path_directory = os.path.join(app_config.path, 'assets')

            if os.path.isdir(path_directory):

                storage = FileSystemStorage(location=path_directory)

                storage.prefix = full_asset_path(app_config.name, "")

                self.locations.append(app_config.name)
                self.storages[app_config.name] = storage

        super().__init__()

    #pylint: disable=redefined-builtin
    def find(self, path, all=False):
        return []

    def list(self, ignore_patterns):
        for component_name in self.locations:
            storage = self.storages[component_name]
            for path in get_files(storage, ignore_patterns + self.ignore_patterns):
                yield path, storage


class DashAssetFinder(BaseFinder):
    'Find static files in asset directories'

    #pylint: disable=unused-import, unused-variable, no-name-in-module, import-error, abstract-method

    def __init__(self):

        # Ensure urls are loaded
        root_urls = settings.ROOT_URLCONF
        importlib.import_module(root_urls)

        # Get all registered django dash apps

        self.apps = all_apps()

        self.locations = []
        self.storages = OrderedDict()

        self.ignore_patterns = ["*.py", "*.pyc",]

        added_locations = {}

        for app_slug, obj in self.apps.items():

            caller_module = obj.caller_module
            location = obj.caller_module_location
            subdir = obj.assets_folder

            path_directory = os.path.join(os.path.dirname(location), subdir)

            if os.path.isdir(path_directory):

                component_name = app_slug
                storage = FileSystemStorage(location=path_directory)
                path = full_asset_path(obj.caller_module.__name__,"")
                storage.prefix = path

                self.locations.append(component_name)
                self.storages[component_name] = storage

        super().__init__()

    #pylint: disable=redefined-builtin
    def find(self, path, all=False):
        return []

    def list(self, ignore_patterns):
        for component_name in self.locations:
            storage = self.storages[component_name]
            for path in get_files(storage, ignore_patterns + self.ignore_patterns):
                yield path, storage




#
# Monkey patching of register_callback
#

import dash._callback

from dash._callback import (handle_grouped_callback_args,
                            insert_callback,
                            NoUpdate,
                            )
import collections
from functools import wraps

from dash.dependencies import (
    handle_callback_args,
    handle_grouped_callback_args,
    Output,
)
from dash.exceptions import PreventUpdate

from dash._grouping import (
    flatten_grouping,
    make_grouping_by_index,
    grouping_len,
)
from dash._utils import (
    create_callback_id,
    stringify_id,
    to_json,
)

from dash import _validate


def register_callback(
    callback_list, callback_map, config_prevent_initial_callbacks, *_args, **_kwargs
):
    (
        output,
        flat_inputs,
        flat_state,
        inputs_state_indices,
        prevent_initial_call,
    ) = handle_grouped_callback_args(_args, _kwargs)
    if isinstance(output, Output):
        # Insert callback with scalar (non-multi) Output
        insert_output = output
        multi = False
    else:
        # Insert callback as multi Output
        insert_output = flatten_grouping(output)
        multi = True

    output_indices = make_grouping_by_index(output, list(range(grouping_len(output))))
    callback_id = insert_callback(
        callback_list,
        callback_map,
        config_prevent_initial_callbacks,
        insert_output,
        output_indices,
        flat_inputs,
        flat_state,
        inputs_state_indices,
        prevent_initial_call,
    )

    # pylint: disable=too-many-locals
    def wrap_func(func):
        @wraps(func)
        def add_context(*args, **kwargs):
            output_spec = kwargs.pop("outputs_list")
            _validate.validate_output_spec(insert_output, output_spec, Output)

            func_args, func_kwargs = _validate.validate_and_group_input_args(
                args, inputs_state_indices
            )

            func_kwargs = {**func_kwargs,
                           **kwargs}

            # don't touch the comment on the next line - used by debugger
            output_value = func(*func_args, **func_kwargs)  # %% callback invoked %%

            if isinstance(output_value, NoUpdate):
                raise PreventUpdate

            if not multi:
                output_value, output_spec = [output_value], [output_spec]
                flat_output_values = output_value
            else:
                if isinstance(output_value, (list, tuple)):
                    # For multi-output, allow top-level collection to be
                    # list or tuple
                    output_value = list(output_value)

                # Flatten grouping and validate grouping structure
                flat_output_values = flatten_grouping(output_value, output)

            _validate.validate_multi_return(
                output_spec, flat_output_values, callback_id
            )

            component_ids = collections.defaultdict(dict)
            has_update = False
            for val, spec in zip(flat_output_values, output_spec):
                if isinstance(val, NoUpdate):
                    continue
                for vali, speci in (
                    zip(val, spec) if isinstance(spec, list) else [[val, spec]]
                ):
                    if not isinstance(vali, NoUpdate):
                        has_update = True
                        id_str = stringify_id(speci["id"])
                        component_ids[id_str][speci["property"]] = vali

            if not has_update:
                raise PreventUpdate

            response = {"response": component_ids, "multi": True}

            try:
                jsonResponse = to_json(response)
            except TypeError:
                _validate.fail_callback_output(output_value, output)

            return jsonResponse

        callback_map[callback_id]["callback"] = add_context

        return func

    return wrap_func


dash._callback.register_callback = register_callback



# Generated by Django 2.0.5 on 2018-06-08 20:02

# pylint: skip-file

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DashApp',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('instance_name', models.CharField(blank=True, max_length=100, unique=True)),
                ('slug', models.SlugField(blank=True, max_length=110, unique=True)),
                ('base_state', models.TextField(default='{}')),
                ('creation', models.DateTimeField(auto_now_add=True)),
                ('update', models.DateTimeField(auto_now=True)),
                ('save_on_change', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='StatelessApp',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('app_name', models.CharField(max_length=100, unique=True)),
                ('slug', models.SlugField(blank=True, max_length=110, unique=True)),
            ],
        ),
        migrations.AddField(
            model_name='dashapp',
            name='stateless_app',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='django_plotly_dash.StatelessApp'),
        ),
    ]



"""Test helpers to configure an app that tests many cases (single/multiple outputs, patterns, multi triggers, ...).


"""

import json
from functools import wraps
from pathlib import Path
from typing import Union

import dash
from dash import html
import flask
from dash import Dash, no_update
from dash.dependencies import MATCH, ALL, ALLSMALLER, Input, Output, State
from dash.development.base_component import Component
from flask import request

from django_plotly_dash import DjangoDash

dash_contract_data = Path(__file__).with_suffix(".json")


def fill_in_test_app(app: Union[DjangoDash, Dash], write=False):
    """Takes an unitialized Dash (or DjangoDash) app and add dash components and callbacks to test
    - single/multiple outputs
    - patterns (ALL,ALLSMALLER,MATCH)
    - multi triggers

    :param app: the Dash/DjangoDash app to initialize with components and callbacks
    :param write: if True, it will record/log in the tests_dash_contract.json all client callbacks to the app
    :return: None
    """

    # define decorator that logs the body/response if write==True
    if not write:
        log_body_response = lambda f: f
    else:
        record_session = dash_contract_data.open("w")
        record_session.write("[")

        def log_body_response(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                result = f(*args, **kwargs)
                record_session.write(
                    json.dumps(
                        {
                            "body": flask.request.get_json(),
                            "args": args,
                            "kwargs": kwargs,
                            "result": result,
                        }
                    )
                )
                record_session.write(",")
                return result

            return wrapper

    def add_outputs_multi():
        """Test output format (in list or not)"""
        inp1 = html.Button("Multi-single-output", id="inp1")
        outs = [html.Div(id=f"out1-{i}") for i in range(4)]

        app.layout.children.append(html.Div([inp1] + outs))

        @app.callback(
            Output(outs[0].id, "children"),
            [Input(inp1.id, "n_clicks_timestamp")],
            [State(inp1.id, "n_clicks")],
        )
        @log_body_response
        def test_single_output(*args):
            return f"single - {args}"

        @app.callback(
            [Output(outs[1].id, "children")],
            [Input(inp1.id, "n_clicks_timestamp")],
            [State(inp1.id, "n_clicks")],
        )
        @log_body_response
        def test_single_output_list(*args):
            return [f"single in list - {args}"]

        @app.callback(
            [Output(outs[2].id, "children"), Output(outs[3].id, "children")],
            [Input(inp1.id, "n_clicks_timestamp")],
            [State(inp1.id, "n_clicks")],
        )
        @log_body_response
        def test_multi_output(*args):
            return [f"multi in list - {args}"] * 2

    def add_update_state_in_output():
        inp1b = html.Button("Input-update-output-state", id="inp1b")
        out1b = html.A(href="http://www.example.com/null", id="out1b")

        app.layout.children.append(html.Div([inp1b, out1b]))

        @app.callback(
            [Output(out1b.id, "href"), Output(out1b.id, "children"), ],
            [Input(inp1b.id, "n_clicks_timestamp")],
        )
        @log_body_response
        def test_update_output_state(n_clicks_timestamp):
            return f"http://www.example.com/{n_clicks_timestamp}", f"http://www.example.com/{n_clicks_timestamp}"


    def add_multi_triggered():
        """Test a callback getting more than one element in the triggered context"""
        inp1 = html.Button("Multiple triggered", id="inp2")
        outs = [html.Div(id=f"out2-{i}") for i in range(1)]

        app.layout.children.append(html.Div([inp1] + outs))

        @app.callback(
            Output(outs[0].id, "children"),
            [Input(inp1.id, "n_clicks_timestamp"), Input(inp1.id, "n_clicks")],
        )
        @log_body_response
        def test_single_output(*args, **kwargs):
            print(kwargs)
            return f"multi triggered - {args} - {dash.callback_context.triggered or []}"

    def add_pattern_all():
        inps = [
            html.Button(f"Pattern ALL {i}", id={"_id": f"inp-{i}", "_type": "btn3"})
            for i in range(3)
        ]
        out = html.Div(id=f"out3")

        app.layout.children.append(html.Div(inps + [out]))

        @app.callback(
            Output(out.id, "children"),
            [Input({"_id": ALL, "_type": "btn3"}, "n_clicks_timestamp")],
            [State({"_id": ALL, "_type": "btn3"}, "n_clicks")],
        )
        @log_body_response
        def test_single_output(*args):
            return f"pattern ALL - {args}"

    def add_pattern_allsmaller():
        inps = [
            html.Button(f"Pattern ALLSMALLER {i}", id={"_id": f"inp-{i}", "_type": "btn4"})
            for i in range(3)
        ]
        out = html.Div(id=f"out4")

        app.layout.children.append(html.Div(inps + [out]))

        @app.callback(
            Output({"_id": MATCH, "_type": "btn4"}, "children"),
            [Input({"_id": ALLSMALLER, "_type": "btn4"}, "n_clicks")],
            [State({"_id": ALLSMALLER, "_type": "btn4"}, "id")],
        )
        @log_body_response
        def test_single_output(*args):
            return f"pattern ALLSMALLER - {args}"

    def add_pattern_match():
        inps = [
            html.Button(f"Pattern MATCH {i}", id={"_id": f"inp-{i}", "_type": "btn5"})
            for i in range(3)
        ]
        out = html.Div(id=f"out5")

        app.layout.children.append(html.Div(inps + [out]))

        @app.callback(
            Output({"_id": MATCH, "_type": "btn5"}, "children"),
            [Input({"_id": MATCH, "_type": "btn5"}, "n_clicks")],
            [State({"_id": MATCH, "_type": "btn5"}, "id")],
        )
        @log_body_response
        def test_single_output(*args):
            return f"pattern MATCH - {args}"

    def add_stop():
        app.layout.children.append(html.Button("stop", id="stop"))

        @app.callback(Output("stop", "children"), [Input("stop", "n_clicks")])
        def stop_server(nclicks):
            if nclicks:
                if write:
                    # close the file recording the session
                    record_session.write("]")
                    record_session.close()
                func = request.environ.get("werkzeug.server.shutdown")
                func()
            return no_update

    app.layout = html.Div([])
    add_outputs_multi()
    add_update_state_in_output()
    add_multi_triggered()
    add_pattern_all()
    add_pattern_allsmaller()
    add_pattern_match()
    add_stop()

    # add to all elements a n_clicks/n_clicks_timestamp values to have populate_values detect properties
    i = 0
    ts = 1611733453854
    for comp in app.layout.children:
        for elem in comp.children:
            if isinstance(elem, Component):
                elem.n_clicks = i
                elem.n_clicks_timestamp = ts
                i += 1
                ts += 100


if __name__ == "__main__":
    # to generate the test_dash_contract.json, run the pure Dash app and click on each button once successively
    # if you regenerate the test_dash_contract.json, the final assert of the test test_dash_stateful_app will fail
    # as the timestamp will be different => the test will have to be updated accordingly
    test_app = Dash("DashContractApp")

    fill_in_test_app(test_app, write=True)

    test_app.run_server(debug=True)



'''
Django-plotly-dash middleware

This middleware enables the collection of items from templates for inclusion in the header and footer

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from .util import serve_locally, static_path

#pylint: disable=too-few-public-methods

class EmbeddedHolder:
    'Hold details of embedded content from processing a view'
    def __init__(self):
        self.css = ""
        self.config = ""
        self.scripts = ""
    def add_css(self, css):
        'Add css content'
        if css:
            self.css = css
    def add_config(self, config):
        'Add config content'
        if config:
            self.config = config
    def add_scripts(self, scripts):
        'Add js content'
        if scripts:
            self.scripts += scripts

class ContentCollector:
    '''
    Collect content during view processing, and substitute in response by finding magic strings.

    This enables view functionality, such as template tags, to introduce content such as css and js
    inclusion into the header and footer.
    '''
    def __init__(self):
        self.header_placeholder = "DJANGO_PLOTLY_DASH_HEADER_PLACEHOLDER"
        self.footer_placeholder = "DJANGO_PLOTLY_DASH_FOOTER_PLACEHOLDER"

        self.embedded_holder = EmbeddedHolder()
        self._encoding = "utf-8"

    def adjust_response(self, response):
        'Locate placeholder magic strings and replace with content'

        try:
            c1 = self._replace(response.content,
                               self.header_placeholder,
                               self.embedded_holder.css)

            response.content = self._replace(c1,
                                             self.footer_placeholder,
                                             "\n".join([self.embedded_holder.config,
                                                        self.embedded_holder.scripts]))
        except AttributeError:
            # Catch the "FileResponse instance has no `content` attribute" error when serving media files in the Django development server.
            pass

        return response

    def _replace(self, content, placeholder, substitution):
        return content.replace(self._encode(placeholder),
                               self._encode(substitution if substitution else ""))

    def _encode(self, string):
        return string.encode(self._encoding)

class BaseMiddleware:
    'Django-plotly-dash middleware'

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):

        request.dpd_content_handler = ContentCollector()
        response = self.get_response(request)
        response = request.dpd_content_handler.adjust_response(response)

        return response


# Bootstrap4 substitutions, if available
dpd_substitutions = []
try:
    from dpd_static_support.mappings import substitutions as dpd_direct_substitutions
    dpd_substitutions += dpd_direct_substitutions
    from dpd_static_support.mappings import static_substitutions as dpd_ss_substitutions
    dpd_substitutions += [(x, static_path(y)) for x, y in dpd_ss_substitutions]
except Exception as e:
    pass


class ExternalRedirectionMiddleware:
    'Middleware to force redirection in third-party content through rewriting'

    def __init__(self, get_response):
        self.get_response = get_response

        substitutions = []

        if serve_locally():
            substitutions += dpd_substitutions

        self._encoding = "utf-8"

        self.substitutions = [(self._encode(source),
                               self._encode(target)) for source, target in substitutions]

    def __call__(self, request):

        response = self.get_response(request)

        try:
            content = response.content

            for source, target in self.substitutions:
                content = content.replace(source, target)

            response.content = content

        except AttributeError:
            # Not all files can contain substitutions, so ignore them
            pass

        return response

    def _encode(self, string):
        return string.encode(self._encoding)




'''
Define pipe consumer and also http endpoint for direct injection of pipe messages

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import json

from channels.generic.websocket import WebsocketConsumer
from channels.generic.http import AsyncHttpConsumer
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync

def _form_pipe_channel_name(channel_name):
    return "dpd_pipe_%s" % channel_name

def send_to_pipe_channel(channel_name,
                         label,
                         value):
    'Send message through pipe to client component'
    async_to_sync(async_send_to_pipe_channel)(channel_name=channel_name,
                                              label=label,
                                              value=value)

async def async_send_to_pipe_channel(channel_name,
                                     label,
                                     value):
    'Send message asynchronously through pipe to client component'
    pcn = _form_pipe_channel_name(channel_name)

    channel_layer = get_channel_layer()
    await channel_layer.group_send(pcn,
                                   {"type":"pipe.value",
                                    "label":label,
                                    "value":value})

class MessageConsumer(WebsocketConsumer):
    'Websocket handler for pipe to dash component'
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.channel_maps = {}

    def connect(self):
        self.accept()

    def disconnect(self, reason):  # pylint: disable=arguments-differ

        for _, pipe_group_name in self.channel_maps.items():
            async_to_sync(self.channel_layer.group_discard)(pipe_group_name, self.channel_name)

        return super().disconnect(reason)

    def pipe_value(self, message):
        'Send a new value into the ws pipe'
        jmsg = json.dumps(message)
        self.send(jmsg)

    def update_pipe_channel(self, uid, channel_name, label): # pylint: disable=unused-argument
        '''
        Update this consumer to listen on channel_name for the js widget associated with uid
        '''
        pipe_group_name = _form_pipe_channel_name(channel_name)

        if self.channel_layer:
            current = self.channel_maps.get(uid, None)
            if current != pipe_group_name:
                if current:
                    async_to_sync(self.channel_layer.group_discard)(current, self.channel_name)

                self.channel_maps[uid] = pipe_group_name
                async_to_sync(self.channel_layer.group_add)(pipe_group_name, self.channel_name)

    def receive(self, text_data): # pylint: disable=arguments-differ
        message = json.loads(text_data)

        message_type = message.get('type', 'unknown_type')
        if message_type == 'connection_triplet':

            try:
                channel_name = message['channel_name']
                uid = message['uid']
                label = message['label']

                self.update_pipe_channel(uid, channel_name, label)

            except: # pylint: disable=bare-except
                # Ignore malformed message
                # TODO enable logging of this sort of thing
                pass

        else:
            # Not a periodic control message, so do something useful
            # For now, this is just pushing to all other pipe consumers indiscrimnately
            # TODO grab defaults from settings if present

            channel_name = message.get('channel_name', "UNNAMED_CHANNEL")
            value = message.get('value', None)
            label = message.get('label', 'DEFAULT$LABEL')

            send_to_pipe_channel(channel_name=channel_name,
                                 label=label,
                                 value=value)

class PokePipeConsumer(AsyncHttpConsumer):
    'Async handling of http request inserting pipe message'

    async def handle(self, body):

        user = self.scope.get('user', None)
        as_utf = body.decode('utf-8')
        try:
            incoming_message = json.loads(as_utf)

            # Get label value and channel_name out of the body
            channel_name = incoming_message.get('channel_name', 'UNNAMED_CHANNEL')
            value = incoming_message.get("value", None)
            label = incoming_message.get("label", "DEFAULT$LABEL")

            # TODO Use user info (and also csrf and other checks as desired) to prevent misuse
            await async_send_to_pipe_channel(channel_name,
                                             label,
                                             value)

            response = """PokePipeConsumer consumed message of %s for %s
""" %(incoming_message, user)
            response_code = 200

        except: # pylint: disable=bare-except

            response = """Unable to understand and forward on message of %s
""" % as_utf
            response_code = 500

        await self.send_response(response_code,
                                 response.encode('utf-8'))



'''
Views for serving up dash applications and their constituents

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

 # pylint: disable=unused-argument

import json

from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect

from dash.exceptions import PreventUpdate

try:
    from dash.fingerprint import check_fingerprint
except:
    # check_fingerprint not available, fake it
    def check_fingerprint(resource):
        return resource, None


from .models import DashApp, check_stateless_loaded
from .util import get_initial_arguments, static_path

def routes(*args, **kwargs):
    'Return routes'
    raise NotImplementedError

def dependencies(request, ident, stateless=False, **kwargs):
    'Return the dependencies'
    _, app = DashApp.locate_item(ident, stateless)

    with app.app_context():
        view_func = app.locate_endpoint_function('dash-dependencies')
        resp = view_func()
        return HttpResponse(resp.data,
                            content_type=resp.mimetype)

def layout(request, ident, stateless=False, cache_id=None, **kwargs):
    'Return the layout of the dash application'
    _, app = DashApp.locate_item(ident, stateless)

    view_func = app.locate_endpoint_function('dash-layout')
    resp = view_func()

    initial_arguments = get_initial_arguments(request, cache_id)

    response_data, mimetype = app.augment_initial_layout(resp, initial_arguments)
    return HttpResponse(response_data,
                        content_type=mimetype)

def update(request, ident, stateless=False, **kwargs):
    try:
        return _update(request, ident, stateless, **kwargs)
    except PreventUpdate:
        return HttpResponse(status=204)

def _update(request, ident, stateless=False, **kwargs):
    'Generate update json response'
    dash_app, app = DashApp.locate_item(ident, stateless)

    request_body = json.loads(request.body.decode('utf-8'))

    # Use direct dispatch with extra arguments in the argMap
    app_state = request.session.get("django_plotly_dash", dict())
    arg_map = {'dash_app_id': ident,
               'dash_app': dash_app,
               'user': request.user,
               'request':request,
               'session_state': app_state}
    resp = app.dispatch_with_args(request_body, arg_map)
    request.session['django_plotly_dash'] = app_state
    dash_app.handle_current_state()

    # Special for ws-driven edge case
    if str(resp) == 'EDGECASEEXIT':
        return HttpResponse("")

    # Change in returned value type
    try:
        rdata = resp.data
        rtype = resp.mimetype
    except:
        rdata = resp
        rtype = "application/json"

    return HttpResponse(rdata,
                        content_type=rtype)

def main_view(request, ident, stateless=False, cache_id=None, **kwargs):
    'Main view for a dash app'
    _, app = DashApp.locate_item(ident, stateless, cache_id=cache_id)

    view_func = app.locate_endpoint_function()
    resp = view_func()
    return HttpResponse(resp)

def component_component_suites(request, resource=None, component=None, **kwargs):
    'Return part of a client-side component, served locally for some reason'
    return component_suites(request,
                            resource=resource,
                            component=component,
                            extra_element="_components/",
                            **kwargs)


def component_suites_build(request, resource=None, component=None, extra_element="", cpe2=None, **kwargs):
    'Return part of a client-side component, served locally for some reason'
    return component_suites(request,
                            resource=resource,
                            component=component,
                            extra_element=extra_element,
                            cpe2=f"{cpe2}/build",
                            **kwargs)


def component_suites(request, resource=None, component=None, extra_element="", cpe2=None, **kwargs):
    'Return part of a client-side component, served locally for some reason'

    extra_path_part = f"{cpe2}/" if cpe2 else ""

    get_params = request.GET.urlencode()
    if get_params and False:
        redone_url = static_path("dash/component/%s/%s%s%s?%s" %(component, extra_path_part, extra_element, resource, get_params))
    else:
        resource, _fingerprint = check_fingerprint(resource)
        redone_url = static_path("dash/component/%s/%s%s%s" % (component, extra_path_part, extra_element, resource))

    return HttpResponseRedirect(redirect_to=redone_url)

def app_assets(request, **kwargs):
    'Return a local dash app asset, served up through the Django static framework'
    get_params = request.GET.urlencode()
    extra_part = ""
    if get_params:
        redone_url = static_path("dash/assets/%s?%s" %(extra_part, get_params))
    else:
        redone_url = static_path("dash/assets/%s" % extra_part)

    return HttpResponseRedirect(redirect_to=redone_url)

# pylint: disable=wrong-import-position, wrong-import-order
from django.template.response import TemplateResponse

def add_to_session(request, template_name="index.html", **kwargs):
    'Add some info to a session in a place that django-plotly-dash can pass to a callback'

    django_plotly_dash = request.session.get("django_plotly_dash", dict())

    session_add_count = django_plotly_dash.get('add_counter', 0)
    django_plotly_dash['add_counter'] = session_add_count + 1
    request.session['django_plotly_dash'] = django_plotly_dash

    return TemplateResponse(request, template_name, {})

def asset_redirection(request, path, ident=None, stateless=False, **kwargs):
    'Redirect static assets for a component'

    X, app = DashApp.locate_item(ident, stateless)

    # Redirect to a location based on the import path of the module containing the DjangoDash app
    static_path = X.get_asset_static_url(path)

    return redirect(static_path)

def add_stateless_apps(request, **kwargs):
    """Check all registered stateless apps and create ORM entries that are missing"""
    check_stateless_loaded()
    return redirect('admin:django_plotly_dash_statelessapp_changelist')




#!/usr/bin/env python

from setuptools import setup


with open('django_plotly_dash/version.py') as f:
    exec(f.read())


with open('README.md') as f:
    long_description = f.read()


setup(
    name="django-plotly-dash",
    version=__version__,
    url="https://github.com/GibbsConsulting/django-plotly-dash",
    description="Django use of plotly dash apps through template tags",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Mark Gibbs",
    author_email="django_plotly_dash@gibbsconsulting.ca",
    license='MIT',
    packages=[
    'django_plotly_dash',
    'django_plotly_dash.migrations',
    'django_plotly_dash.templatetags',
    ],
    include_package_data=True,
    classifiers = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Framework :: Dash',
    ],
    keywords='django plotly plotly-dash dash dashboard',
    project_urls = {
    'Source': "https://github.com/GibbsConsulting/django-plotly-dash",
    'Tracker': "https://github.com/GibbsConsulting/django-plotly-dash/issues",
    'Documentation': 'http://django-plotly-dash.readthedocs.io/',
    },
    install_requires = ['plotly',
                        'dash>=2.0',
                        'dpd-components',

                        'dash-bootstrap-components',

                        'channels>=2.0',
                        'Django>=2.2,<4.0.0',
                        'Flask>=1.0.2',
                        'Werkzeug',
    ],
    python_requires=">=3.8",
    )




'''
Tests of django_plotly_dash.

These tests can be directly used or imported into eg the tests of the demo app.

The use of pytest is assumed.

Copyright (c) 2018 Gibbs Consulting and others - see CONTRIBUTIONS.md

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

import json
from unittest.mock import patch

import pytest
# pylint: disable=bare-except
from dash.dependencies import Input, State, Output
from django.urls import reverse

from django_plotly_dash import DjangoDash
from django_plotly_dash.dash_wrapper import get_local_stateless_list, get_local_stateless_by_name
from django_plotly_dash.models import DashApp, find_stateless_by_name
from django_plotly_dash.tests_dash_contract import fill_in_test_app, dash_contract_data


def test_dash_app():
    'Test the import and formation of the dash app orm wrappers'

    from django_plotly_dash.models import StatelessApp
    stateless_a = StatelessApp(app_name="Some name")

    assert stateless_a
    assert stateless_a.app_name
    assert str(stateless_a) == stateless_a.app_name


@pytest.mark.django_db
def test_dash_stateful_app_client_contract(client):
    'Test the state management of a DashApp as well as the contract between the client and the Dash app'

    from django_plotly_dash.models import StatelessApp

    # create a DjangoDash, StatelessApp and DashApp
    ddash = DjangoDash(name="DDash")
    fill_in_test_app(ddash, write=False)

    stateless_a = StatelessApp(app_name="DDash")
    stateless_a.save()
    stateful_a = DashApp(stateless_app=stateless_a,
                         instance_name="Some name",
                         slug="my-app", save_on_change=True)
    stateful_a.save()

    # check app can be found back
    assert "DDash" in get_local_stateless_list()
    assert get_local_stateless_by_name("DDash") == ddash
    assert find_stateless_by_name("DDash") == ddash

    # check the current_state is empty
    assert stateful_a.current_state() == {}

    # set the initial expected state
    expected_state = {'inp1': {'n_clicks': 0, 'n_clicks_timestamp': 1611733453854},
                      'inp1b': {'n_clicks': 5, 'n_clicks_timestamp': 1611733454354},
                      'inp2': {'n_clicks': 7, 'n_clicks_timestamp': 1611733454554},
                      'out1-0': {'n_clicks': 1, 'n_clicks_timestamp': 1611733453954},
                      'out1-1': {'n_clicks': 2, 'n_clicks_timestamp': 1611733454054},
                      'out1-2': {'n_clicks': 3, 'n_clicks_timestamp': 1611733454154},
                      'out1-3': {'n_clicks': 4, 'n_clicks_timestamp': 1611733454254},
                      'out1b': {'href': 'http://www.example.com/null',
                                'n_clicks': 6,
                                'n_clicks_timestamp': 1611733454454},
                      'out2-0': {'n_clicks': 8, 'n_clicks_timestamp': 1611733454654},
                      'out3': {'n_clicks': 12, 'n_clicks_timestamp': 1611733455054},
                      'out4': {'n_clicks': 16, 'n_clicks_timestamp': 1611733455454},
                      'out5': {'n_clicks': 20, 'n_clicks_timestamp': 1611733455854},
                      '{"_id":"inp-0","_type":"btn3"}': {'n_clicks': 9,
                                                         'n_clicks_timestamp': 1611733454754},
                      '{"_id":"inp-0","_type":"btn4"}': {'n_clicks': 13,
                                                         'n_clicks_timestamp': 1611733455154},
                      '{"_id":"inp-0","_type":"btn5"}': {'n_clicks': 17,
                                                         'n_clicks_timestamp': 1611733455554},
                      '{"_id":"inp-1","_type":"btn3"}': {'n_clicks': 10,
                                                         'n_clicks_timestamp': 1611733454854},
                      '{"_id":"inp-1","_type":"btn4"}': {'n_clicks': 14,
                                                         'n_clicks_timestamp': 1611733455254},
                      '{"_id":"inp-1","_type":"btn5"}': {'n_clicks': 18,
                                                         'n_clicks_timestamp': 1611733455654},
                      '{"_id":"inp-2","_type":"btn3"}': {'n_clicks': 11,
                                                         'n_clicks_timestamp': 1611733454954},
                      '{"_id":"inp-2","_type":"btn4"}': {'n_clicks': 15,
                                                         'n_clicks_timestamp': 1611733455354},
                      '{"_id":"inp-2","_type":"btn5"}': {'n_clicks': 19,
                                                         'n_clicks_timestamp': 1611733455754}}

    ########## test state management of the app and conversion of components ids
    # search for state values in dash layout
    stateful_a.populate_values()
    assert stateful_a.current_state() == expected_state
    assert stateful_a.have_current_state_entry("inp1", "n_clicks")
    assert stateful_a.have_current_state_entry({"_type": "btn3", "_id": "inp-0"}, "n_clicks_timestamp")
    assert stateful_a.have_current_state_entry('{"_id":"inp-0","_type":"btn3"}', "n_clicks_timestamp")
    assert not stateful_a.have_current_state_entry("checklist", "other-prop")

    # update a non existent state => no effect on current_state
    stateful_a.update_current_state("foo", "value", "random")
    assert stateful_a.current_state() == expected_state

    # update an existent state => update current_state
    stateful_a.update_current_state('{"_id":"inp-2","_type":"btn5"}', "n_clicks", 100)
    expected_state['{"_id":"inp-2","_type":"btn5"}'] = {'n_clicks': 100, 'n_clicks_timestamp': 1611733455754}
    assert stateful_a.current_state() == expected_state

    assert DashApp.objects.get(instance_name="Some name").current_state() == {}

    stateful_a.handle_current_state()

    assert DashApp.objects.get(instance_name="Some name").current_state() == expected_state

    # check initial layout serve has the correct values injected
    dash_instance = stateful_a.as_dash_instance()
    resp = dash_instance.serve_layout()

    # initialise layout with app state
    layout, mimetype = dash_instance.augment_initial_layout(resp, {})
    assert '"n_clicks": 100' in layout

    # initialise layout with initial arguments
    layout, mimetype = dash_instance.augment_initial_layout(resp, {
        '{"_id":"inp-2","_type":"btn5"}': {"n_clicks": 200}})
    assert '"n_clicks": 100' not in layout
    assert '"n_clicks": 200' in layout

    ########### test contract between client and app by replaying interactions recorded in tests_dash_contract.json
    # get update component route
    url = reverse('the_django_plotly_dash:update-component', kwargs={'ident': 'my-app'})

    # for all interactions in the tests_dash_contract.json
    for scenario in json.load(dash_contract_data.open("r")):
        body = scenario["body"]

        response = client.post(url, json.dumps(body), content_type="application/json")

        assert response.status_code == 200

        response = json.loads(response.content)

        # compare first item in response with first result
        result = scenario["result"]
        if isinstance(result, list):
            result = result[0]
        content = response["response"].popitem()[1].popitem()[1]
        assert content == result

        # handle state
        stateful_a.handle_current_state()

    # check final state has been changed accordingly
    final_state = {'inp1': {'n_clicks': 1, 'n_clicks_timestamp': 1615103027288},
                   'inp1b': {'n_clicks': 5, 'n_clicks_timestamp': 1615103033482},
                   'inp2': {'n_clicks': 8, 'n_clicks_timestamp': 1615103036591},
                   'out1-0': {'n_clicks': 1, 'n_clicks_timestamp': 1611733453954},
                   'out1-1': {'n_clicks': 2, 'n_clicks_timestamp': 1611733454054},
                   'out1-2': {'n_clicks': 3, 'n_clicks_timestamp': 1611733454154},
                   'out1-3': {'n_clicks': 4, 'n_clicks_timestamp': 1611733454254},
                   'out1b': {'href': 'http://www.example.com/1615103033482',
                             'n_clicks': 6,
                             'n_clicks_timestamp': 1611733454454},
                   'out2-0': {'n_clicks': 8, 'n_clicks_timestamp': 1611733454654},
                   'out3': {'n_clicks': 12, 'n_clicks_timestamp': 1611733455054},
                   'out4': {'n_clicks': 16, 'n_clicks_timestamp': 1611733455454},
                   'out5': {'n_clicks': 20, 'n_clicks_timestamp': 1611733455854},
                   '{"_id":"inp-0","_type":"btn3"}': {'n_clicks': 10,
                                                      'n_clicks_timestamp': 1615103039030},
                   '{"_id":"inp-0","_type":"btn4"}': {'n_clicks': 14,
                                                      'n_clicks_timestamp': 1611733455154},
                   '{"_id":"inp-0","_type":"btn5"}': {'n_clicks': 18,
                                                      'n_clicks_timestamp': 1611733455554},
                   '{"_id":"inp-1","_type":"btn3"}': {'n_clicks': 11,
                                                      'n_clicks_timestamp': 1615103039496},
                   '{"_id":"inp-1","_type":"btn4"}': {'n_clicks': 15,
                                                      'n_clicks_timestamp': 1611733455254},
                   '{"_id":"inp-1","_type":"btn5"}': {'n_clicks': 19,
                                                      'n_clicks_timestamp': 1611733455654},
                   '{"_id":"inp-2","_type":"btn3"}': {'n_clicks': 12,
                                                      'n_clicks_timestamp': 1615103040528},
                   '{"_id":"inp-2","_type":"btn4"}': {'n_clicks': 15,
                                                      'n_clicks_timestamp': 1611733455354},
                   '{"_id":"inp-2","_type":"btn5"}': {'n_clicks': 20,
                                                      'n_clicks_timestamp': 1611733455754}}

    assert DashApp.objects.get(instance_name="Some name").current_state() == final_state



def test_dash_callback_arguments():
    'Test the flexibility of the callback arguments order (handling of inputs/outputs/states)'

    # create a DjangoDash
    ddash = DjangoDash(name="DashCallbackArguments")

    # add a callback with the new flexible order of dependencies
    @ddash.callback(
        Output("one", "foo"),
        Output("two", "foo"),
        Input("one", "baz"),
        Input("two", "baz"),
        Input("three", "baz"),
        State("one", "bil"),
    )
    def new():
        pass

    # add a callback with the old/classical flexible order of dependencies
    @ddash.callback(
        [Output("one", "foo"),
         Output("two", "foo")],
        [Input("one", "baz"),
         Input("two", "baz"),
         Input("three", "baz")],
        [State("one", "bil")]
    )
    def old():
        pass

    assert ddash._callback_sets == [({'inputs': [Input("one", "baz"),
                                                 Input("two", "baz"),
                                                 Input("three", "baz"), ],
                                      'output': [Output("one", "foo"),
                                                 Output("two", "foo")],
                                      'prevent_initial_call': None,
                                      'state': [State("one", "bil"), ]},
                                     new),
                                    ({'inputs': [Input("one", "baz"),
                                                 Input("two", "baz"),
                                                 Input("three", "baz"), ],
                                      'output': [Output("one", "foo"),
                                                 Output("two", "foo")],
                                      'prevent_initial_call': None,
                                      'state': [State("one", "bil"), ]},
                                     old)
                                    ]


def test_util_error_cases(settings):
    'Test handling of missing settings'

    settings.PLOTLY_DASH = None

    from django_plotly_dash.util import pipe_ws_endpoint_name, dpd_http_endpoint_root, http_endpoint, insert_demo_migrations

    assert pipe_ws_endpoint_name() == 'dpd/ws/channel'
    assert dpd_http_endpoint_root() == "dpd/views"
    assert http_endpoint("fred") == '^dpd/views/fred/$'
    assert not insert_demo_migrations()

    del settings.PLOTLY_DASH

    assert pipe_ws_endpoint_name() == 'dpd/ws/channel'
    assert dpd_http_endpoint_root() == "dpd/views"
    assert http_endpoint("fred") == '^dpd/views/fred/$'
    assert not insert_demo_migrations()


def test_demo_routing():
    'Test configuration options for the demo'

    from django_plotly_dash.util import pipe_ws_endpoint_name, insert_demo_migrations
    assert pipe_ws_endpoint_name() == 'ws/channel'
    assert insert_demo_migrations()


def test_local_serving(settings):
    'Test local serve settings'

    from django_plotly_dash.util import serve_locally, static_asset_root, full_asset_path
    assert serve_locally() == settings.DEBUG
    assert static_asset_root() == 'dpd/assets'
    assert full_asset_path('fred.jim', 'harry') == 'dpd/assets/fred/jim/harry'


@pytest.mark.django_db
def test_direct_access(client):
    'Check direct use of a stateless application using demo test data'

    from django.urls import reverse
    from .app_name import main_view_label

    for route_name in ['layout', 'dependencies', main_view_label]:
        for prefix, arg_map in [('app-', {'ident':'SimpleExample'}),
                                ('', {'ident':'simpleexample-1'}),]:
            url = reverse('the_django_plotly_dash:%s%s' % (prefix, route_name), kwargs=arg_map)

            response = client.get(url)

            assert response.content
            assert response.status_code == 200

    for route_name in ['routes',]:
        for prefix, arg_map in [('app-', {'ident':'SimpleExample'}),
                                ('', {'ident':'simpleexample-1'}),]:
            url = reverse('the_django_plotly_dash:%s%s' % (prefix, route_name), kwargs=arg_map)

            did_fail = False
            try:
                response = client.get(url)
            except:
                did_fail = True

            assert did_fail


@pytest.mark.django_db
def test_updating(client):
    'Check updating of an app using demo test data'

    from django.urls import reverse

    route_name = 'update-component'

    for prefix, arg_map in [('app-', {'ident':'SimpleExample'}),
                            ('', {'ident':'simpleexample-1'}),]:
        url = reverse('the_django_plotly_dash:%s%s' % (prefix, route_name), kwargs=arg_map)

        response = client.post(url, json.dumps({'output': 'output-size.children',
                                                'inputs':[{'id':'dropdown-color',
                                                           'property':'value',
                                                           'value':'blue'},
                                                          {'id':'dropdown-size',
                                                           'property':'value',
                                                           'value':'medium'},
                                                         ]}), content_type="application/json")

        assert response.content == b'{"response":{"output-size":{"children":"The chosen T-shirt is a medium blue one."}},"multi":true}'
        assert response.status_code == 200


@pytest.mark.django_db
def test_injection_app_access(client):
    'Check direct use of a stateless application using demo test data'

    from django.urls import reverse
    from .app_name import main_view_label

    for route_name in ['layout', 'dependencies', main_view_label]:
        for prefix, arg_map in [('app-', {'ident':'dash_example_1'}),
                                #('', {'ident':'simpleexample-1'}),
                               ]:
            url = reverse('the_django_plotly_dash:%s%s' % (prefix, route_name), kwargs=arg_map)

            response = client.get(url)

            assert response.content
            assert response.status_code == 200

    for route_name in ['routes',]:
        for prefix, arg_map in [('app-', {'ident':'dash_example_1'}),]:
            url = reverse('the_django_plotly_dash:%s%s' % (prefix, route_name), kwargs=arg_map)

            did_fail = False
            try:
                response = client.get(url)
            except:
                did_fail = True

            assert did_fail



@pytest.mark.django_db
def test_injection_updating_multiple_callbacks(client):
    'Check updating of an app using demo test data for multiple callbacks'

    from django.urls import reverse

    route_name = 'update-component'

    for prefix, arg_map in [('app-', {'ident':'multiple_callbacks'}),]:
        url = reverse('the_django_plotly_dash:%s%s' % (prefix, route_name), kwargs=arg_map)

        # output is now a string of id and propery
        response = client.post(url, json.dumps({'output':'..output-one.children...output-two.children...output-three.children..',
                                                'inputs':[
            {'id':'button',
             'property':'n_clicks',
             'value':'10'},
            {'id':'dropdown-color',
             'property':'value',
             'value':'purple-ish yellow with a hint of greeny orange'},
                                                         ]}), content_type="application/json")

        assert response.status_code == 200

        resp = json.loads(response.content.decode('utf-8'))
        assert 'response' in resp

        resp_detail = resp['response']
        assert 'output-two' in resp_detail
        assert 'children' in resp_detail['output-two']
        assert resp_detail['output-two']['children'] == "Output 2: 10 purple-ish yellow with a hint of greeny orange []"


@pytest.mark.django_db
def test_flexible_expanded_callbacks(client):
    'Check updating of an app using demo test data for flexible expanded callbacks'

    from django.urls import reverse

    route_name = 'update-component'

    for prefix, arg_map in [('app-', {'ident':'flexible_expanded_callbacks'}),]:
        url = reverse('the_django_plotly_dash:%s%s' % (prefix, route_name), kwargs=arg_map)

        # output contains all arguments of the expanded_callback
        response = client.post(url, json.dumps({'output':'output-one.children',
                                                'inputs':[
            {'id':'button',
             'property':'n_clicks',
             'value':'10'},
                                                         ]}), content_type="application/json")

        assert response.status_code == 200

        resp = json.loads(response.content.decode('utf-8'))
        for key in ["dash_app_id", "dash_app", "callback_context"]:
            assert key in resp["response"]['output-one']['children']

        # output contains all arguments of the expanded_callback
        response = client.post(url, json.dumps({'output':'output-two.children',
                                                'inputs':[
            {'id':'button',
             'property':'n_clicks',
             'value':'10'},
                                                         ]}), content_type="application/json")

        assert response.status_code == 200

        resp = json.loads(response.content.decode('utf-8'))
        assert resp["response"]=={'output-two': {'children': 'ok'}}


        # output contains all arguments of the expanded_callback
        response = client.post(url, json.dumps({'output':'output-three.children',
                                                'inputs':[
            {'id':'button',
             'property':'n_clicks',
             'value':'10'},
                                                         ]}), content_type="application/json")

        assert response.status_code == 200

        resp = json.loads(response.content.decode('utf-8'))
        assert resp["response"]=={"output-three": {"children": "flexible_expanded_callbacks"}}


@pytest.mark.django_db
def test_injection_updating(client):
    'Check updating of an app using demo test data'

    from django.urls import reverse

    route_name = 'update-component'

    for prefix, arg_map in [('app-', {'ident':'dash_example_1'}),]:
        url = reverse('the_django_plotly_dash:%s%s' % (prefix, route_name), kwargs=arg_map)

        response = client.post(url, json.dumps({#'output':{'id':'test-output-div', 'property':'children'},
                                                'output': "test-output-div.children",
                                                'inputs':[{'id':'my-dropdown1',
                                                           'property':'value',
                                                           'value':'TestIt'},
                                                         ]}), content_type="application/json")

        rStart = b'{"response":{"test-output-div":{"children":[{"props":{"id":"line-area-graph2"'

        assert response.content.startswith(rStart)
        assert response.status_code == 200

        # Single output callback, output=="component_id.component_prop"
        response = client.post(url, json.dumps({'output':'test-output-div.children',
                                                'inputs':[{'id':'my-dropdown1',
                                                           'property':'value',
                                                           'value':'TestIt'},
                                                         ]}), content_type="application/json")

        rStart = b'{"response":{"test-output-div":{"children":[{"props":{"id":"line-area-graph2"'

        assert response.content.startswith(rStart)
        assert response.status_code == 200

        # Single output callback, fails if output=="..component_id.component_prop.."
        with pytest.raises(KeyError, match="..test-output-div.children.."):
            client.post(url, json.dumps({'output':'..test-output-div.children..',
                                                'inputs':[{'id':'my-dropdown1',
                                                           'property':'value',
                                                           'value':'TestIt'},
                                                         ]}), content_type="application/json")


        # Multiple output callback, fails if output=="component_id.component_prop"
        with pytest.raises(KeyError, match="test-output-div3.children"):
            client.post(url, json.dumps({'output':'test-output-div3.children',
                                                'inputs':[{'id':'my-dropdown1',
                                                           'property':'value',
                                                           'value':'TestIt'},
                                                         ]}), content_type="application/json")

        # Multiple output callback, output=="..component_id.component_prop.."
        response = client.post(url, json.dumps({'output':'..test-output-div3.children..',
                                                'inputs':[{'id':'my-dropdown1',
                                                           'property':'value',
                                                           'value':'TestIt'},
                                                         ]}), content_type="application/json")

        rStart = b'{"response":{"test-output-div3":{"children":[{"props":{"id":"line-area-graph2"'

        assert response.content.startswith(rStart)
        assert response.status_code == 200

        with pytest.raises(KeyError, match="django_to_dash_context"):
            client.post(url, json.dumps({'output': 'test-output-div2.children',
                                         'inputs':[{'id':'my-dropdown2',
                                                    'property':'value',
                                                    'value':'TestIt'},
                                                  ]}), content_type="application/json")

        session = client.session
        session['django_plotly_dash'] = {'django_to_dash_context': 'Test 789 content'}
        session.save()

        response = client.post(url, json.dumps({'output': 'test-output-div2.children',
                                                 'inputs':[{'id':'my-dropdown2',
                                                            'property':'value',
                                                            'value':'TestIt'},
                                                          ]}), content_type="application/json")
        rStart = b'{"response":{"test-output-div2":{"children":[{"props":{"children":["You have '

        assert response.content.startswith(rStart)
        assert response.status_code == 200

        assert response.content.find(b'Test 789 content') > 0


@pytest.mark.django_db
def test_argument_settings(settings, client):
    'Test the setting that controls how initial arguments are propagated through to the dash app'

    from django_plotly_dash.util import initial_argument_location, store_initial_arguments, get_initial_arguments

    assert initial_argument_location()

    settings.PLOTLY_DASH = {'cache_arguments': True}

    assert initial_argument_location()

    test_value = {"test":"first"}

    cache_id = store_initial_arguments(None, test_value)

    assert len(cache_id) > 10

    fetched = get_initial_arguments(None, cache_id)

    assert fetched == test_value

    settings.PLOTLY_DASH = {'cache_arguments': False}

    assert not initial_argument_location()

    cache_id2 = store_initial_arguments(client, test_value)

    assert len(cache_id2) > 10

    assert cache_id != cache_id2

    ## For some reason, sessions are continually replaced, so lookup here doesnt work
    #fetched2 = get_initial_arguments(client, cache_id2)
    #assert fetched2 == test_value

    assert store_initial_arguments(None, None) is None
    assert get_initial_arguments(None, None) is None
    assert store_initial_arguments(client, None) is None
    assert get_initial_arguments(client, None) is None


def test_stateless_lookup_noop():
    'Test no-op stateless lookup'

    from django_plotly_dash.util import stateless_app_lookup_hook
    lh_hook = stateless_app_lookup_hook()

    assert lh_hook is not None
    with pytest.raises(ImportError):
        lh_hook("not an app")


def test_middleware_artifacts():
    'Import and vaguely exercise middleware objects'

    from django_plotly_dash.middleware import EmbeddedHolder, ContentCollector

    eh = EmbeddedHolder()
    eh.add_css("some_css")
    eh.add_config("some_config")
    eh.add_scripts("some_scripts")

    assert eh.config == 'some_config'

    cc = ContentCollector()

    assert cc._encode("fred") == b'fred'


def test_finders():
    'Import and vaguely exercise staticfiles finders'

    from django_plotly_dash.finders import DashComponentFinder, DashAppDirectoryFinder, DashAssetFinder

    dcf = DashComponentFinder()
    dadf = DashAppDirectoryFinder()
    daf = DashAssetFinder()

    assert dcf is not None
    assert dadf is not None
    assert daf is not None


@pytest.mark.django_db
def test_app_loading(client):

    from django_plotly_dash.models import check_stateless_loaded
    from django.urls import reverse

    # Function should run wthout raising errors
    check_stateless_loaded()
    assert True

    url = reverse('the_django_plotly_dash:add_stateless_apps')

    response = client.post(url)

    # This view redirects to the main admin
    assert response.status_code == 302


@pytest.mark.django_db
def test_external_scripts_stylesheets(client):
    'Check external_stylesheets and external_scripts ends up in index'

    from demo.plotly_apps import external_scripts_stylesheets
    dash = external_scripts_stylesheets.as_dash_instance()

    with patch.object(dash, "interpolate_index") as mock:
        dash.index()

    _, kwargs = mock.call_args
    assert "https://codepen.io/chriddyp/pen/bWLwgP.css" in kwargs["css"]
    assert "https://stackpath.bootstrapcdn.com/bootstrap/4.1.3/css/bootstrap.min.css" in kwargs["css"]
    assert "https://www.google-analytics.com/analytics.js" in kwargs["scripts"]
    assert "https://cdn.polyfill.io/v2/polyfill.min.js" in kwargs["scripts"]
    assert "https://cdnjs.cloudflare.com/ajax/libs/lodash.js/4.17.10/lodash.core.js" in kwargs["scripts"]

def test_callback_decorator():
    inputs = [Input("one", "value"),
              Input("two", "value"),
              ]
    states = [Input("three", "value"),
              Input("four", "value"),
              ]

    def callback_standard(one, two, three, four):
        return

    assert DjangoDash.get_expanded_arguments(callback_standard, inputs, states) == []

    def callback_standard(one, two, three, four, extra_1):
        return

    assert DjangoDash.get_expanded_arguments(callback_standard, inputs, states) == ['extra_1']

    def callback_args(one, *args):
        return

    assert DjangoDash.get_expanded_arguments(callback_args, inputs, states) == []

    def callback_args_extra(one, *args, extra_1):
        return

    assert DjangoDash.get_expanded_arguments(callback_args_extra, inputs, states) == ['extra_1' ]

    def callback_args_extra_star(one, *, extra_1):
        return

    assert DjangoDash.get_expanded_arguments(callback_args_extra_star, inputs, states) == ['extra_1' ]


    def callback_kwargs(one, two, three, four, extra_1, **kwargs):
        return

    assert DjangoDash.get_expanded_arguments(callback_kwargs, inputs, states) == None

    def callback_kwargs(one, two, three, four, *, extra_1, **kwargs, ):
        return

    assert DjangoDash.get_expanded_arguments(callback_kwargs, inputs, states) == None
